import random
import math
from typing import Generator, List, TypeVar

T = TypeVar("T")


def random_iter(values: List[T]) -> Generator[T, None, None]:
    """Yield a random value picked from the provided list of values, forever.

    Args:
        values: A list of possible values to yield.

    Yields:
        A value randomly chosen from the list of values.
    """
    while True:
        yield random.choice(values)


def random_sequences(sequences: List[List[T]]) -> Generator[List[T], None, None]:
    """Yield values from a sequence randomly chosen from a list of sequences, forever.

    Args:
        sequences: A list of lists of values (a.k.a. sequences).

    Yields:
        The next value in the current randomly chosen sequence. Once that sequence is
        exhausted, another sequence will be randomly chosen from the provided list.
    """
    while True:
        fragment = random.choice(sequences)
        for value in fragment:
            yield value


def coin_flip(p: float) -> bool:
    """Flip a biased coin.

    Args:
        p: probability between 0 and 1 of the coin landing heads up.

    Returns:
        True if the coin lands heads up, False otherwise.
    """
    return p > random.random()


def every_n_ticks(n: int, tick: int, offset: int = 0) -> bool:
    """Fire a signal every n ticks.

    Args:
        n: the number of ticks to wait before signalling.
        tick: the current tick number.
        offset: the amount by which to shift the provided tick number.

    Returns:
        True if n ticks have passed, False otherwise.
    """
    return not (tick + offset) % n


def major_scale(root: int) -> List[int]:
    """Build a MIDI major scale from the given root.

    Args:
        root: the MIDI note number to use as the scale's root.

    Returns:
        A list of 7 MIDI notes representing a major scale with the given root.
    """
    assert (
        0 <= root and root < 127
    ), f"the provided root is not a valid MIDI note: {root}"
    return [root, root + 2, root + 4, root + 5, root + 7, root + 9, root + 11]


def maybe_octave(odds: int) -> Generator[int, None, None]:
    """Return either 0 or 12, with variable odds.

    Args:
        odds: the odds of returning 0 instead of 12, e.g., an odds of 5
              will yield 0 five times more often than 12.

    Yields:
        0 or 12, with 0 likelier than 12 according to `odds`.
    """
    assert odds >= 0, f"expected non-negative odds but received: {odds}"
    return random_iter([0] * odds + [12])


def sine(tick: int, min: int, max: int, period: int, shift: int = 0) -> int:
    """A sine wave that is a function of the current tick.

    Args:
        tick: the current tick number.
        min: the minimum value of the sine wave.
        max: the maximum value of the sine wave.
        period: the period of the sine wave in ticks.
        shift: the tick offset of the sine wave.

    Returns:
        The value of the sine wave at the provided tick number.
    """
    x = (tick - shift) / period * 2 * math.pi
    y = math.sin(x) * 0.5 + 0.5
    return int(y * (max - min) + min)


def incrat(tick: int, scale: float):
    """A function of the provided tick whose value asymptotically approaches 1 as tick increases.

    Args:
        tick: the current tick number.
        scale: a positive number for modifying the rate of increase of the function.

    Returns:
        The value of the function for the given tick.
    """
    scaled_tick = tick * scale
    return scaled_tick / (scaled_tick + 1)


# So you don't have to import the python random module separately
r = random


import functools
import asyncio
from typing import Callable, List, NoReturn, Optional, Sequence

import mido


class MidiController:
    """Trigger note and other CC parameter events on a particular MIDI channel.

    Use as a standalone or connect to a `Sequencer` instance via the `Sequencer.register` method.

    Args:
        channel: the MIDI channel to send messages on.
    """

    def __init__(self, channel: int):
        self.channel = channel
        self.port = None
        self.note_tasks = {}

    def _set_port(self, port: mido.ports.BaseOutput):
        self.port = port

    def reset(self):
        """Send `note_off` messages for any currently playing notes."""
        for note, task in self.note_tasks.items():
            task.cancel()
            self.note_off(note)

    def note_off(self, note: int):
        """Send a `note_off` MIDI message with the given `note` value."""
        self.port.send(mido.Message("note_off", note=note, channel=self.channel))

    def note_on(self, note: int, velocity: int):
        """Send a `note_on` MIDI message with the given `note` and `velocity` values."""
        self.port.send(
            mido.Message("note_on", note=note, velocity=velocity, channel=self.channel)
        )

    def set_cc(self, control: int, value: int):
        """Send a MIDI message to CC `control` with the given `value`."""
        self.port.send(
            mido.Message(
                "control_change", control=control, value=value, channel=self.channel
            )
        )

    async def play_note(self, note: int, velocity: int, duration: float):
        """Play a `note` at a `velocity` for a `duration` (in seconds)."""
        if note in self.note_tasks:
            self.note_tasks[note].cancel()
            self.note_off(note)
        self.note_on(note, velocity)
        await asyncio.sleep(duration)
        self.note_off(note)


Generator = Callable[[MidiController, int], NoReturn]


class Sequencer:
    """Provide time data to sets of MidiControllers, triggering time-based events.

    Args:
        output_device: the name of the MIDI output device to use.
        tps: ticks per second, how many time events are fired per second.

    Examples:
        >>> s = Sequencer()
        >>> @s.register([MidiController(1), MidiController(2)])
        ... async def music_generator(ctrl, tick):
        ...     ctrl.play_note(60, 60, 1)

    Todos:
        Shift from using ticks per second to configure the sequencer's speed
        to beats per minute and pulses per beat. This allows for configuring speed
        and rhythmic resolution separately, via BPM and PPB respectively.
    """

    def __init__(self, output_device: str = "IAC Driver Bus 1", tps=100):
        self.tps = tps
        self.spt = 1 / tps
        self.port = mido.open_output(output_device, autoreset=True)
        self.generators: List[Generator] = []
        self.controllers: List[MidiController] = []
        self.tick = 1

    def register(self, controllers: List[MidiController]):
        """Decorator for an asynchronous function that takes a `MidiController` and
        the current `tick` as its arguments.

        Every time this `Sequencer` instance's `tick` value increases, it will call
        this function with the updated tick value, once for each `MidiController` in
        the provided list of `controllers`.
        """
        self.controllers.extend(controllers)

        for ctrl in controllers:
            ctrl._set_port(self.port)

        def decorator(generator: Generator):
            @functools.wraps(generator)
            async def wrapped(tick: int):
                for ctrl in controllers:
                    await generator(ctrl, tick)

            self.generators.append(wrapped)

            return wrapped

        return decorator

    async def _loop(self):
        while True:
            for subscriber in self.generators:
                asyncio.create_task(subscriber(self.tick))
                await asyncio.sleep(self.spt)
            self.tick += 1

    def play(self):
        """Run the sequencer, starting from a tick value of 1."""
        if self.generators == []:
            raise Exception("sequencer has no generators")

        try:
            asyncio.run(self._loop())
        except KeyboardInterrupt:
            for controller in self.controllers:
                controller.reset()
            self.tick = 1



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lucier",
    version="0.0.2",
    author="Jacob Lurye",
    description="Sequence MIDI messages using a Flask-style interface.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jacoblurye/lucier",
    packages=["lucier"],
    install_requires=["mido==1.2.9", "python-rtmidi==1.4.6"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
